/**
 * 
 */
package com.platomix.userprofile.utils;

/**
 * @author wookvn
 *
 */
public final class StrBuilder {

	private StringBuilder sqlBuilder = new StringBuilder();
	
	public StrBuilder() {}
	
	public StrBuilder(String value) {
		sqlBuilder.append(value);
	}
	
	public StrBuilder append(String value) {
		sqlBuilder.append(value);
		return this;
	}
	
	public StrBuilder appendLn(String value) {
		sqlBuilder.append(value).append("\n");
		return this;
	}
	
	public StrBuilder deleteLastChar() {
		sqlBuilder = sqlBuilder.deleteCharAt(sqlBuilder.length() - 1);
		return this;
	}
	
	@Override
	public String toString() {
		return sqlBuilder.toString();
	}
	
}